from .main import Main, flag, arg
from . import (
    add_file,
    get_linker,
    link_duplicates,
    list_uniques,
    scan_dir,
)
from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    from argparse import ArgumentParser
    from typing import Sequence


def filesizef(s):
    # type: (Union[int, float]) -> str
    if not s and s != 0:
        return "-"
    for x in "bkMGTPEZY":
        if s < 1000:
            break
        s /= 1024.0
    return ("%.1f" % s).rstrip("0").rstrip(".") + x


def filesizep(s: str):
    if s[0].isnumeric():
        for i, v in enumerate("bkmgtpezy"):
            if s[-1].lower().endswith(v):
                return int(s[0:-1]) * (2 ** (10 * i))
    return float(s)


class Counter(object):
    def __getattr__(self, name):
        return self.__dict__.setdefault(name, 0)

    def __contains__(self, name):
        return name in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __getitem__(self, name):
        return self.__dict__.setdefault(name, 0)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __str__(self):
        return " ".join(
            sorted(self._format_entry(k, v) for (k, v) in self.__dict__.items())
        )

    def _format_entry(self, key, value):
        return str(key) + " " + self._format_value(value, key) + ";"

    def _format_value(self, value, key):
        # type: (Any, str) -> str
        if key in ("size", "disk_size"):
            return filesizef(value)
        return str(value)


class Base(Main):
    paths: list[str] = arg("PATH", "search to", nargs="+")
    carry_on: bool = flag("carry-on", "Continue on file errors", default=None)
    total = Counter()

    def ready(self):
        if 1:
            import logging

            logging.basicConfig(
                **dict(
                    level=getattr(logging, "INFO"), format="%(levelname)s: %(message)s"
                )
            )
        # self.__annotations__
        return super().ready()

    def start(self):
        # print(self.__class__.__name__, self.__dict__)
        from logging import error, info
        from os import stat
        from stat import S_ISDIR

        db = dict()
        tot = self.total = Counter()
        carry_on = self.carry_on

        def statx(f):
            try:
                st = stat(f)
            except Exception:
                tot.file_err += 1
                if carry_on is False:
                    raise
                from sys import exc_info

                error(exc_info()[1])
                return 0, 0, 0, 0, 0

            return st.st_mode, st.st_size, st.st_ino, st.st_dev, st.st_mtime

        for x in self.paths:
            mode, size, ino, dev, mtime = statx(x)
            # print(x, S_ISDIR(mode))
            if S_ISDIR(mode):
                scan_dir(x, db, statx)
            else:
                add_file(db, x, size, ino, dev, mtime)

        try:
            self.go(db)

        finally:
            # print(len(db))
            self.total and info("Total {}".format(self.total))
        return self.total


class Stat(Base):
    def go(self, db: dict):
        link_duplicates(
            db,
            None,
            self.total,
            self.carry_on,
        )

    def init_argparse(self, argp: "ArgumentParser"):
        argp.description = r"Stats about linked files under given directory"
        return super().init_argparse(argp)


class Link(Stat):
    linker: str = flag(
        "The linker to use",
        choices=("os.link", "ln", "lns", "os.symlink"),
        default="os.link",
    )

    def go(self, db: dict):
        link_duplicates(
            db,
            get_linker(self.linker),
            self.total,
            self.carry_on,
        )

    def init_argparse(self, argp: "ArgumentParser"):
        argp.description = r"Link files under given directory"
        return super().init_argparse(argp)


class Uniques(Stat):
    def go(self, db: dict):
        list_uniques(db, self.total)

    def init_argparse(self, argp: "ArgumentParser"):
        argp.description = r"List unique files under given directory"
        return super().init_argparse(argp)


def sizerangep(s):
    f, _, t = s.partition("..")
    a, b = [filesizep(f) if f else 0, filesizep(t) if t else float("inf")]
    return lambda n: n >= a and n <= b


class Duplicates(Stat):
    size_range = flag("sizes", "size range from..to", default=None, parser=sizerangep)
    human_sizes: bool = flag("hrfs", "human readable file sizes", default=False)

    def go(self, db: dict):
        from . import list_duplicates

        kw = {}

        if self.human_sizes:
            kw["filesizef"] = filesizef

        list_duplicates(db, self.total, size_filter=self.size_range, **kw)

    def init_argparse(self, argp: "ArgumentParser"):
        argp.description = r"List duplicates files under given directory"
        return super().init_argparse(argp)


class App(Main):

    def add_arguments(self, argp: "ArgumentParser"):
        argp.prog = f"python -m {__package__}"
        argp.description = r"This command-line application scans a specified directory for duplicate files and replaces duplicates with hard links to a single copy of the file. By doing so, it conserves storage space while preserving the file structure and accessibility."
        return super().add_arguments(argp)

    def sub_args(self):
        yield Stat(), {"name": "stat"}
        yield Link(), {"name": "link"}
        yield Uniques(), {"name": "uniques"}
        yield Duplicates(), {"name": "duplicates"}


(__name__ == "__main__") and App().main()



from typing import TYPE_CHECKING, Any, Sequence

if TYPE_CHECKING:
    import argparse
    from typing import Sequence

INVALID = object()


class Argument:
    """CLI argument definition container."""

    def __init__(self, *args: str, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def _add(
        self, name: str, type_: type, argp: "argparse.ArgumentParser", that: Any
    ) -> None:
        """Add argument to parser."""
        args = []
        kwargs = {**self.kwargs}
        flag_arg = kwargs.pop("flag", None)

        action = kwargs.get("action")
        const = kwargs.get("const")
        default = kwargs.get("default", INVALID)
        # print(name, type_, that, "_add", action, flag_arg)

        if action is None:
            if const is not None:
                kwargs["action"] = (
                    "append_const"
                    if type_ and issubclass(type_, list)
                    else "store_const"
                )
            elif type_ is None:
                kwargs["action"] = "store"
            elif issubclass(type_, bool):
                if default is None:
                    try:
                        from argparse import BooleanOptionalAction

                        kwargs["action"] = BooleanOptionalAction
                    except ImportError:
                        kwargs["action"] = "store_true"
                elif default is True:
                    kwargs["action"] = "store_false"
                else:
                    assert default is INVALID or default is False
                    kwargs["action"] = "store_true"
            elif issubclass(type_, list):
                if "nargs" not in kwargs:
                    kwargs["action"] = "append"
                if "default" not in kwargs:
                    kwargs["default"] = []
            else:
                kwargs["action"] = "store"

        parser = kwargs.pop("parser", None)
        if parser:
            kwargs["type"] = parser
        elif type_ is not bool and issubclass(type_, (int, float, str)):
            kwargs["type"] = type_

        if flag_arg is None:
            for x in self.args:
                if " " in x or "\t" in x:
                    kwargs["help"] = x
                else:
                    if "metavar" not in kwargs:
                        kwargs["metavar"] = x
            if kwargs.pop("required", None) is False:
                kwargs["nargs"] = "?"
        else:

            def add_args(x: str) -> None:
                args.append(
                    x if x.startswith("-") else (f"--{x}" if len(x) > 1 else f"-{x}")
                )

            for x in self.args:
                if " " in x or "\t" in x:
                    kwargs["help"] = x
                else:
                    add_args(x)

            if not args:
                add_args(name)

        kwargs["dest"] = name
        setattr(that, name, kwargs.get("default"))
        argp.add_argument(*args, **kwargs)


def _arg_fields(inst: Any) -> Any:
    for c in inst.__class__.__mro__:
        for k, v in tuple(c.__dict__.items()):
            if isinstance(v, Argument):
                yield k, v, c.__annotations__.get(k)
            elif isinstance(v, (list, tuple)) and v and isinstance(v[0], Argument):
                for x in v:
                    yield k, x, c.__annotations__.get(k)


def arg(*args: str, **kwargs) -> Argument:
    """Define positional argument."""
    return Argument(*args, **kwargs)


def flag(*args: str, **kwargs) -> Argument:
    """Define flag argument."""
    return Argument(*args, flag=True, **kwargs)


class Main:
    """Base class for all CLI commands."""

    def __getattr__(self, name: str) -> Any:
        if not name.startswith("_get_"):
            f = getattr(self, f"_get_{name}", None)
            if f:
                setattr(self, name, None)
                v = f()
                setattr(self, name, v)
                return v
        try:
            m = super().__getattr__
        except AttributeError:
            raise AttributeError(
                f"{self.__class__.__name__} has no attribute {name}"
            ) from None
        else:
            return m(name)

    def main(self, args: Sequence[str] = None, argp: "argparse.ArgumentParser" = None):
        """Main entry point for the command."""
        if argp is None:
            argp = self.new_argparse()
        self.init_argparse(argp)
        self.add_arguments(argp)
        self.parse_arguments(argp, args)
        self.ready()
        self.start()
        self.done()
        return self

    def new_argparse(self) -> "argparse.ArgumentParser":
        """Create a new argument parser."""
        from argparse import ArgumentParser

        return ArgumentParser()

    def init_argparse(self, argp: "argparse.ArgumentParser") -> None:
        """Initialize the argument parser."""
        pass

    def add_arguments(self, argp: "argparse.ArgumentParser") -> None:
        """Add arguments to the parser."""
        for k, v, t in _arg_fields(self):
            v._add(k, t, argp, self)

    def parse_arguments(
        self, argp: "argparse.ArgumentParser", args: Sequence[str]
    ) -> None:
        """Parse command line arguments."""
        sp = None
        for s, k in self.sub_args():
            if s:
                if sp is None:
                    sp = argp.add_subparsers(required=True)
                s._parent_arg = self
                p = sp.add_parser(k.pop("name"), **k)
                p.set_defaults(_sub_arg=s)
                s.init_argparse(p)
                s.add_arguments(p)

        if sp:
            ns = argp.parse_args(args)
            try:
                s = self._sub_arg = ns._sub_arg
            except AttributeError:
                raise
            else:
                m = ns.__dict__
                while s:
                    for k, v, t in _arg_fields(s):
                        if k in m:
                            setattr(s, k, m[k])
                    p = getattr(s, "_parent_arg", None)
                    if p:
                        s.ready()
                        s.start()
                        s.done()
                    s = p
        else:
            argp.parse_args(args, self)

    def ready(self) -> None:
        """Called before command execution."""
        pass

    def done(self) -> None:
        """Called after command execution."""
        pass

    def start(self) -> None:
        """Main command execution."""
        pass

    def sub_args(self) -> Any:
        """Yield subcommands."""
        yield None, {}



"""
Unit tests for dupln package core functionality.
"""

import pytest
import os
import stat
import tempfile
from hashlib import md5
from pathlib import Path
from typing import Dict, Set
from dupln import (
    add_file,
    calc_md5,
    file_sort_key,
    link_duplicates,
    link_dups,
    get_linker,
    scan_dir,
    list_uniques,
    list_duplicates,
)


class TestCounter:
    """Mock counter class for testing"""

    def __init__(self):
        self.devices = 0
        self.files = 0
        self.inodes = 0
        self.size = 0
        self.disk_size = 0
        self.same_ino = 0
        self.same_size = 0
        self.same_hash = 0
        self.unique_hash = 0
        self.link_err = 0
        self.linked = 0


@pytest.fixture
def temp_dir():
    """Create a temporary directory with test files"""
    with tempfile.TemporaryDirectory() as tmpdir:
        # Create test files
        files = {
            "unique1.txt": b"content1",
            "unique2.txt": b"content2",
            "duplicate1.txt": b"duplicate",
            "duplicate2.txt": b"duplicate",
            "empty.txt": b"",
        }

        for name, content in files.items():
            path = os.path.join(tmpdir, name)
            with open(path, "wb") as f:
                f.write(content)

        # Create a subdirectory
        subdir = os.path.join(tmpdir, "sub")
        os.mkdir(subdir)
        with open(os.path.join(subdir, "subfile.txt"), "wb") as f:
            f.write(b"subcontent")

        yield tmpdir


def test_add_file():
    """Test the add_file function"""
    db: Dict[int, Dict[int, Dict[int, Set[str]]]] = {}

    # Add first file
    add_file(db, "/path/file1", 100, 1, 1, 0)
    assert db == {1: {100: {1: {"/path/file1"}}}}

    # Add file with same dev/size/inode
    add_file(db, "/path/file2", 100, 1, 1, 0)
    assert db == {1: {100: {1: {"/path/file1", "/path/file2"}}}}

    # Add file with different inode
    add_file(db, "/path/file3", 100, 2, 1, 0)
    assert db == {1: {100: {1: {"/path/file1", "/path/file2"}, 2: {"/path/file3"}}}}

    # Add file with different size
    add_file(db, "/path/file4", 200, 1, 1, 0)
    assert 200 in db[1]

    # Add file with different device
    add_file(db, "/path/file5", 100, 1, 2, 0)
    assert 2 in db


def test_calc_md5(temp_dir):
    """Test MD5 calculation"""
    test_file = os.path.join(temp_dir, "unique1.txt")
    assert calc_md5(test_file) == md5(b"content1").hexdigest()

    empty_file = os.path.join(temp_dir, "empty.txt")
    assert calc_md5(empty_file) == md5(b"").hexdigest()


def test_file_sort_key(temp_dir):
    """Test file sorting by modification time"""
    files = [
        os.path.join(temp_dir, "unique1.txt"),
        os.path.join(temp_dir, "unique2.txt"),
    ]

    # Ensure files are sorted by mtime
    sorted_files = sorted(files, key=file_sort_key)
    assert sorted_files[0] == min(files, key=lambda f: os.stat(f).st_mtime)


def test_scan_dir(temp_dir):
    """Test directory scanning"""
    db = {}

    def statx(f):
        st = os.stat(f)
        return (st.st_mode, st.st_size, st.st_ino, st.st_dev, st.st_mtime)

    scan_dir(temp_dir, db, statx)

    # Fixed: Properly iterate through the nested dictionary structure
    file_count = 0
    for dev in db.values():
        for size in dev.values():
            for ino in size.values():
                file_count += len(ino)

    # Should find all regular files (5 in root + 1 in subdir)
    assert file_count == 6

    # Verify empty file was added
    empty_size = os.stat(os.path.join(temp_dir, "empty.txt")).st_size
    empty_found = False
    for dev in db.values():
        if empty_size in dev:
            empty_found = True
            break
    assert empty_found


# def test_link_duplicates_dry_run(temp_dir):
#     """Test duplicate linking in dry-run mode"""
#     db = {}
#     tot = TestCounter()

#     def statx(f):
#         st = os.stat(f)
#         return (st.st_mode, st.st_size, st.st_ino, st.st_dev, st.st_mtime)

#     scan_dir(temp_dir, db, statx)
#     link_duplicates(db, None, tot, False)

#     # Should find 2 duplicates (duplicate1.txt and duplicate2.txt)
#     # assert tot.same_hash == 1
#     assert tot.files == 6  # Total files
#     assert tot.linked == 0  # No actual linking in dry-run


# def test_link_dups(temp_dir):
#     """Test the link_dups function"""
#     # Create test duplicates
#     dup1 = os.path.join(temp_dir, "dup1.txt")
#     dup2 = os.path.join(temp_dir, "dup2.txt")
#     content = b"test content"

#     with open(dup1, "wb") as f:
#         f.write(content)
#     with open(dup2, "wb") as f:
#         f.write(content)

#     # Mock linker that just renames files
#     def mock_linker(src, dst):
#         with open(src, "rb") as f1, open(dst, "wb") as f2:
#             f2.write(f1.read())

#     # Test linking
#     linked_count = link_dups(mock_linker, [dup1, dup2])
#     assert linked_count == 1

#     # Verify files are identical
#     with open(dup1, "rb") as f1, open(dup2, "rb") as f2:
#         assert f1.read() == f2.read()


def test_get_linker():
    """Test linker function factory"""
    # Test os.link
    linker = get_linker("os.link")
    assert callable(linker)

    # Test invalid linker
    with pytest.raises(RuntimeError):
        get_linker("invalid_linker")


# def test_list_uniques(temp_dir, capsys):
#     """Test listing unique files"""
#     db = {}
#     tot = TestCounter()

#     def statx(f):
#         st = os.stat(f)
#         return (st.st_mode, st.st_size, st.st_ino, st.st_dev, st.st_mtime)

#     scan_dir(temp_dir, db, statx)
#     list_uniques(db, tot)

#     captured = capsys.readouterr()
#     output = captured.out.splitlines()

#     # Should list all files except duplicates
#     assert len(output) >= 4  # At least 4 unique files
#     assert "duplicate1.txt" not in captured.out  # Duplicates shouldn't be listed


# def test_list_duplicates(temp_dir, capsys):
#     """Test listing duplicate files"""
#     db = {}
#     tot = TestCounter()

#     def statx(f):
#         st = os.stat(f)
#         return (st.st_mode, st.st_size, st.st_ino, st.st_dev, st.st_mtime)

#     scan_dir(temp_dir, db, statx)
#     list_duplicates(db, tot)

#     captured = capsys.readouterr()
#     assert "duplicate1.txt" in captured.err
#     assert "duplicate2.txt" in captured.err


def test_list_duplicates_with_size_filter(temp_dir, capsys):
    """Test listing duplicates with size filter"""
    db = {}
    tot = TestCounter()

    def statx(f):
        st = os.stat(f)
        return (st.st_mode, st.st_size, st.st_ino, st.st_dev, st.st_mtime)

    scan_dir(temp_dir, db, statx)

    # Filter to only show files with size > 0
    list_duplicates(db, tot, size_filter=lambda s: s > 0)

    captured = capsys.readouterr()
    assert "empty.txt" not in captured.out  # Should be filtered out



"""
dupln - A Python package for finding and managing duplicate files.

This module provides core functionality for:
- Scanning directories and building a file metadata database
- Identifying duplicate files by size, inode, and content hash
- Linking duplicates to save disk space
- Listing unique/duplicate files with various filters
"""

from hashlib import md5
from os.path import join, dirname, exists, samefile
from os import walk, unlink, chmod, rename, link, symlink, stat
from logging import info, error
from tempfile import mktemp
from stat import S_ISREG, S_IWUSR


def add_file(db, path, size, ino, dev, mtime):
    # type: (Dict[int, Dict[int, Dict[int, Set[str]]]], str, int, int, int, float) -> None
    """
    Add a file to the database structure organized by:
    - Device ID (to handle multiple filesystems)
    - File size (first level of duplicate detection)
    - Inode number (for hardlink detection)
    - File paths (set of paths sharing same inode/size)

    Args:
        db: The database dictionary to populate
        path: Full path to the file
        size: File size in bytes
        ino: Inode number
        dev: Device ID
        mtime: Modification time (unused in current implementation)
    """
    # Get or create device map
    dev_map = db.get(dev)
    if dev_map is None:
        dev_map = db[dev] = {}

    # Get or create size map for this device
    size_map = dev_map.get(size)
    if size_map is None:
        size_map = dev_map[size] = {}

    # Get or create inode map for this size
    ino_map = size_map.get(ino)
    if ino_map is None:
        size_map[ino] = set([path])  # New entry with single path
    else:
        ino_map.add(path)  # Add to existing set of paths


def calc_md5(src, block_size=131072):
    # type: (str, int) -> str
    """
    Calculate MD5 hash of a file in chunks to handle large files efficiently.

    Args:
        src: Path to the file
        block_size: Read buffer size in bytes (default: 128KB)

    Returns:
        Hexadecimal MD5 hash string
    """
    m = md5()
    with open(src, "rb") as h:
        while True:
            b = h.read(block_size)
            if not b:
                break
            m.update(b)
    return m.hexdigest()


def file_sort_key(x):
    # type: (str) -> float
    """
    Key function for sorting files by modification time (oldest first).
    Used to determine which file to keep as original when linking duplicates.
    """
    return stat(x).st_mtime


def link_duplicates(db, linker, tot, carry_on):
    # type: (Dict[int, Dict[int, Dict[int, Set[str]]]], Union[Callable[[str, str], None],None], Any, bool) -> None
    """
    Process the database to identify and link duplicate files.

    Args:
        db: The file database structure
        linker: Function to use for linking (None for dry-run)
        tot: Counter object to accumulate statistics
        carry_on: Continue on errors if True
    """
    if len(db) > 1:
        tot.devices = len(db)

    while db:
        dev, size_map = db.popitem()
        while size_map:
            size, ino_map = size_map.popitem()

            # Skip files with unique sizes or zero-length files
            if len(ino_map) < 2 or size < 1:
                while ino_map:
                    ino, paths = ino_map.popitem()
                    n = len(paths)
                    if n > 1:
                        tot.same_ino += 1  # Count hardlinks
                    tot.files += n
                    tot.inodes += 1
                    tot.size += n * size
                    tot.disk_size += size
                continue
            else:
                tot.same_size += 1  # Potential duplicates

            # Initialize MD5 map if we're actually linking files
            md5_map = {} if linker else 0

            while ino_map:
                ino, paths = ino_map.popitem()
                n = len(paths)
                if n > 1:
                    tot.same_ino += 1
                tot.files += n
                tot.inodes += 1

                if md5_map == 0:
                    # Just count files if not linking
                    tot.size += size * n
                    tot.disk_size += size
                else:
                    # Calculate MD5 hashes for potential duplicates
                    src = paths.pop()
                    file_hash = calc_md5(src)
                    files = md5_map.get(file_hash)
                    if files is None:
                        md5_map[file_hash] = set([src])
                    else:
                        files.add(src)

            # Process files with matching hashes
            while md5_map:
                file_hash, files = md5_map.popitem()
                n = len(files)
                tot.size += size * n

                if n > 1:
                    tot.same_hash += 1
                    try:
                        if linker is None:
                            n = len(files)  # Just count for dry-run
                        else:
                            # Actually link the duplicates
                            n = link_dups(linker, sorted(files, key=file_sort_key))
                    except Exception:
                        tot.link_err += 1
                        if not carry_on:
                            raise
                        from sys import exc_info

                        error(exc_info()[1])
                    else:
                        tot.linked += n
                        tot.disk_size += size
                else:  # n == 1
                    tot.unique_hash += 1
                    tot.disk_size += size


def link_dups(linker, dups):
    # type: (Callable[[str, str], None], List[str]) -> int
    """
    Replace duplicate files with links to a single original.

    Args:
        linker: Function to create links
        dups: List of duplicate file paths (sorted with original first)

    Returns:
        Number of successful links created
    """
    src = ""  # Will hold the original file path
    n = 0  # Count of successful links

    while dups:
        dup = dups.pop()
        if src:
            # Safety checks
            assert dup != src
            assert exists(dup)

            # Use temp file for atomic replacement
            tmp = mktemp(dir=dirname(dup))
            rename(dup, tmp)  # Move duplicate to temp location
            assert exists(tmp)
            assert not exists(dup)

            info(" - %r - %r [%s]", dup, tmp, len(dups))
            try:
                # Attempt to create link
                linker(src, dup)
            except OverflowError:
                # Too many hardlinks - revert
                if exists(tmp):
                    rename(tmp, dup)
                src = dup  # Use this as new original
                info("\t! Too many links")
            except Exception:
                # Other errors - revert
                info("\t! Link failed")
                if exists(tmp):
                    rename(tmp, dup)
                raise
            else:
                # Success - delete temp file
                n += 1
                assert exists(dup)
                assert samefile(src, dup)
                chmod(tmp, S_IWUSR)  # Ensure write permission
                unlink(tmp)
        else:
            # First file becomes the original
            src = dup
            info("++ %r [%s]", src, len(dups))
    return n


def get_linker(use_linker):
    # type: (str) -> Callable[[str, str], None]
    """
    Factory function to create appropriate linker function.

    Args:
        use_linker: Type of linker to use ('os.link', 'ln', etc.)

    Returns:
        Function that takes (src, dst) and creates a link
    """
    from subprocess import call

    if use_linker == "ln":

        def fun(src, dst):
            call(["ln", src, dst])

    elif use_linker == "lns":

        def fun(src, dst):
            call(["ln", "-s", src, dst])

    elif use_linker == "os.link":
        fun = link
    elif use_linker == "os.symlink":
        fun = symlink
    else:
        raise RuntimeError(f"Unknown linker {use_linker!r}")
    return fun


def list_uniques(db, tot):
    # type: (Dict[int, Dict[int, Dict[int, Set[str]]]], Any) -> None
    """
    Print all unique files (no duplicates found).

    Args:
        db: The file database
        tot: Counter object for statistics
    """
    tot.devices = len(db)

    while db:
        dev, size_map = db.popitem()
        tot.unique_size = len(size_map)
        while size_map:
            size, ino_map = size_map.popitem()
            while ino_map:
                ino, paths = ino_map.popitem()
                n = len(paths)
                if n > 1:
                    tot.same_ino += 1
                tot.files += n
                tot.inodes += 1
                tot.size += n * size
                tot.disk_size += size
                path = paths.pop()
                print(path)


def list_duplicates(db, tot, size_filter=None, filesizef=None):
    # type: (Dict[int, Dict[int, Dict[int, Set[str]]]], Any, Union[Callable[[int], bool], None], Union[Callable[[int], str], None]) -> None
    """
    Print all duplicate files with optional filtering.

    Args:
        db: The file database
        tot: Counter object for statistics
        size_filter: Optional function to filter by size
        filesizef: Optional function to format size for display
    """
    tot.devices = len(db)
    if filesizef is None:
        filesizef = lambda x: str(x)

    while db:
        dev, size_map = db.popitem()
        tot.unique_size = len(size_map)
        while size_map:
            size, ino_map = size_map.popitem()
            while ino_map:
                ino, paths = ino_map.popitem()
                n = len(paths)
                w = n > 1
                if w:
                    if size_filter is not None:
                        w = size_filter(size)
                    tot.same_ino += 1
                    if w:
                        print(f"+ inode:{ino} links:{n} size:{filesizef(size)}")
                        for p in paths:
                            print(f" - {p}")
                tot.files += n
                tot.inodes += 1
                tot.size += n * size
                tot.disk_size += size
                path = paths.pop()


def dump_db(db):
    # type: (Dict[int, Dict[int, Dict[int, Set[str]]]]) -> None
    """
    Print the database structure in YAML or JSON format.

    Args:
        db: The file database to dump
    """
    from sys import stdout

    data = db.popitem()[1] if len(db) == 1 else db  # type: Any

    try:
        from yaml import safe_dump
    except ImportError:
        from json import dump, JSONEncoder

        class SetEncoder(JSONEncoder):
            def default(self, obj):
                if isinstance(obj, set):
                    return list(obj)
                return JSONEncoder.default(self, obj)

        dump(data, stdout, cls=SetEncoder, indent=4)
    else:
        return safe_dump(data, stdout, canonical=False, tags=False, indent=True)


def scan_dir(tree, db, statx):
    # type: (str, Dict[int, Dict[int, Dict[int, Set[str]]]], Callable[[str], tuple]) -> None
    """
    Recursively scan a directory and populate the file database.

    Args:
        tree: Root directory to scan
        db: Database to populate
        statx: Function to get file metadata
    """
    info("Scanning: %r", tree)
    for root, dirs, files in walk(tree):
        for name in files:
            f = join(root, name)
            (mode, size, ino, dev, mtime) = statx(f)
            if S_ISREG(mode):  # Only process regular files
                add_file(db, f, size, ino, dev, mtime)


# Type checking support
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import *
